/* CGI_mode.cpp
 Dr. R. Wallace 

 Several modifications by Anthony Taylor <taylor@cs.mwsu.edu> to work
 with the alice-core and to make life easier. :) */

#include "alice.h"

int main() {
  /* This program detects the environment variable QUERY_STRING. */
  /* QUERY_STRING has a format like:                             */
  /*     text=Hi+there&submit=Reply&virtual=none                 */
  char client_id[16];    /* unique client ID    */ 
  char timestamp[16];    /* based on timestamp  */
  char datafilename[32]; /* parameter file      */
  char cookietime[42];   /* time cookie expires */
  char *default_query = new char[4096];
  char *scrap = new char[4096];
  char *alice_output;
  struct tm* ltime; 
  time_t clock; 
  char *pos;

  /* initialize the alice-core */
  if (initialize("data/alice.ini")) return(1);

  /* Find out what time it is, we use this for the client_id */
  /*   and for setting our cookie.                           */
  time( &clock );
  ltime = gmtime( &clock );

  /* Get HTTP input: */
  char *http_input = getenv("QUERY_STRING");
  if (http_input == NULL || strlen(http_input) < 1) {
    sprintf(default_query, "text=&submit=Reply&virtual=none");
    http_input = default_query;
  }
  replace(http_input, "text=", "");
  replace(http_input, "&submit=Reply", "");

  /* this next section extracts the client_id (virtual IP)  */
  pos = strstr(http_input, "&virtual=")+9; 
  if (pos != NULL && strlen(pos) < 16) sprintf(client_id, "%s", pos);
  else sprintf(client_id, "%s", "none");

  /* Generate a new ID for a new client. */
  /* Any "new" client has client_id=none: */
  /* if the "new client has none they might */
  /* have a cookie, so we check for that.  */

  if (!strcmp(client_id, "none")) {
    sprintf(client_id,"%s",getenv("HTTP_COOKIE"));
    replace(client_id,"VIR=", "");
    if (strstr(client_id,"=") || !strcmp(client_id,"(null)")) {
      sprintf(timestamp, "%02d%02d%02d%02d%02d",
      ltime->tm_mon,ltime->tm_mday,ltime->tm_hour,
      ltime->tm_min,ltime->tm_sec);
      sprintf(client_id, "%s",timestamp);
    }
  }

  /* HTTP cleanup phase: */

  replace(http_input, "&virtual=none", "");
  replace(http_input, "&virtual=", "");     
  replace(http_input, client_id, ""); 
#define REPLACE(x, y, z) while(replace(x, y, z))
  REPLACE(http_input, "+", " ");
  REPLACE(http_input, "%3F", "?");
  REPLACE(http_input, "%2C", ",");      
  REPLACE(http_input, "%21", ".");
  REPLACE(http_input, "%22", " ");
  REPLACE(http_input, "%27", "'");
  REPLACE(http_input, "%3A", ":");
  REPLACE(http_input, "%3B", ";");
  REPLACE(http_input, "%28", "(");
  REPLACE(http_input, "%29", ")");
  REPLACE(http_input, "%7B", "{");
  REPLACE(http_input, "%7D", "}");
  REPLACE(http_input, "%26", "&");
  REPLACE(http_input, "%5E", "^");
  REPLACE(http_input, "%40", "@");
  REPLACE(http_input, "%C0", "");
  REPLACE(http_input, "%C1", "");
  REPLACE(http_input, "%C2", ""); 
  REPLACE(http_input, "%C3", "");
  REPLACE(http_input, "%C4", "");
  REPLACE(http_input, "%3E", ">");
  REPLACE(http_input, "%3C", "<");

  /* Assume the directory "log" exist: */
  sprintf(datafilename,"log/%s.txt", client_id);

  /* Remember this client's data: */
  loadvars( datafilename ); 
  /* set the ip for logging */
  setvar("ip",getenv("REMOTE_ADDR"));

  alice_output = respond( http_input );    


  /* OUTPUT HTML: */
  printf("Content-type: text/html\n");

  /* Set our cookie */
  ltime->tm_mon += 1;  /* cookie expires one month from now */
  strftime(cookietime, 42, "%A, %d-%b-%y %X %Z", ltime);
  printf("Set-Cookie: VIR=%s; expires=%s; path=/;\n",client_id,cookietime);
  printf("\n");

  /* open up skeleton.txt this has a "skeleton" of what we want our */
  /* HTML to look like.                                             */
  FILE *skel = fopen("data/skeleton.txt", "r");
  fgets(scrap, 4096, skel);
  while(!feof(skel)){
    REPLACE(scrap,"<client_id>",client_id);
    REPLACE(scrap,"<client_input>",http_input);
    REPLACE(scrap,"<alice_output>",alice_output);
    while ((pos = strstr(scrap,"<getvar"))){
      pos = strstr(pos, "\"") + 1;
      strcpy(default_query,pos);
      *strstr(default_query,"\"") = 0;
      REPLACE(scrap,"<get",getvar(default_query,NULL));
      strremove(scrap,"var name","/>");
    }

    printf("%s",scrap);
    fgets(scrap,4096,skel);
  }

  savevars( datafilename ); /* datafilename set above */
  deinitialize();

  return 0; /* we are done with the cgi-script */
}













