/*
--Program C: an implementation of ALICE and AIML in C.
--Copyright (C) 2000 Jacco Bikker
--This program is free software; you can redistribute it and/or modify
--it under the terms of the GNU General Public License as published by
--the Free Software Foundation.     
*/

/* 
--This program has been modified significantly since Jacco's
--original release, please see THANKS in the text directory 
--for contributions.
*/

#include "alice.h"

/* star holds "*" and "_", that holds our that, pbuffer is sort of 
a general scrap variable, although it does eventually hold our
answer. */

char *star, *that, *pbuffer; 

/* best holds the pattern matched */
char *best; 

/* var's for our init file. */
FILE *aimlfile, *errorfile;
FILE *substitutefile, *templatefile, *patternfile;
FILE *logfile;
FILE *gossipfile, *personffile, *personfile, *person2file;
char *datafile, *unsortedfile;

/* should we log? */
int logging;
int log_error;

/* new variables for the getvar setvar code. */
var_array *root_var;

fpos_t chunk[28];

void alierror(char* text){
  if (log_error) fprintf(errorfile,text);
  else printf(text);
}

char *get_pattern_matched(){
  return(best);
}

int setvar(int protect, char *varname, char *value) {
  var_array *conductor, *one_before;
  lowercase(varname);
  if(root_var == NULL) {
    root_var = new var_array;
    root_var->varname = new char[MAX_VARNAME_SIZE];
    root_var->value = new char[MAX_VARVAL_SIZE];
    root_var->protect = protect;
    strcpy(root_var->varname,varname);
    if(value[0] == '*'){ 
      root_var->def = 1;
      strcpy(root_var->value,value + 1);
    }
    else {
      strcpy(root_var->value,value);
      root_var->def = 0;
    }
    root_var->next = NULL;
    if (root_var->def == 0) return(1);
    return(2);
  } 
  conductor = root_var;
  one_before = root_var;

  /* we search through the array for the name */
  while(conductor != NULL) {
    if (!(strcmp(varname,conductor->varname))){
      if((protect == 1) || (conductor->protect == protect)){
        strcpy(conductor->value, value);
      }
      else return(0);
      if (conductor->def == 0) return(1);
          return(2);
    }
    one_before = conductor;
    conductor = conductor->next;
  }
  /* if we get here, the variable doesn't exsist. */
  /* create a new entry. */
  one_before->next = new var_array;
  one_before = one_before->next;
  one_before->varname = new char[MAX_VARNAME_SIZE];
  one_before->value = new char[MAX_VARVAL_SIZE];
  one_before->protect = protect;
  strcpy(one_before->varname,varname);
  if(value[0] == '*'){ 
    one_before->def = 1;
    strcpy(one_before->value,value + 1);
  }
  else {
    strcpy(one_before->value,value);
    one_before->def = 0;
  }
  one_before->next = NULL;
  if (one_before->def == 0) return(1);
    return(2);     
}  

char *getvar(char *varname, char *vardefault) {
  char *value = NULL;
  var_array *conductor;
  lowercase(varname);
  if (strlen(varname) > MAX_VARNAME_SIZE) return(value);
  /* we begin searching through the linked list. */
  conductor = root_var;
  while(conductor != NULL) {
    if (!strcmp(varname,conductor->varname)) {
      value = conductor->value;
      return(value);
    }
    conductor = conductor->next;
  }  
  /* if we made it to here, the variable doesn't exsist. */
  if (vardefault == NULL) return (value);
  if (strlen(vardefault) > MAX_VARVAL_SIZE) return (value);
  setvar(0,varname,vardefault);
  return (vardefault);
}

char *respond(char *text){
  /* Thanks to Dr. Wallace for writing this into the CGI,
     it was stripped out of there, modified slightly, 
     then added to here. */
  char *scrap = new char[MAX_LINE_SIZE];
  char *ltimestr;
  char *uip;
  char *oneline;
  char *output;
  struct tm *ltime;
  time_t clock;

  time(&clock);
  ltime = localtime(&clock); 
  ltimestr = asctime(ltime);

  strcpy(scrap, text);
  substitute(substitutefile,scrap);
  /* perhaps some code to deal with remaining periods that are not
     sentence breaks.  (Like 5.02 or H.I.P.P.i.e.) */
  while(replace(scrap,"?","."));
  while(replace(scrap,"!","."));
  while(replace(scrap,"\n",""));
  /* we need to make sure our sentence ends in a period. */
  if(scrap[strlen(scrap)-2] != '.'){
    strcat(scrap,".");  
  }
  uip = scrap;
  oneline = new char[MAX_LINE_SIZE];
  output = new char[MAX_LINE_SIZE];
  output[0] = 0;
  do {    
    memset(oneline,0, MAX_LINE_SIZE);
    /* skip white space:*/
    while(uip[0] == ' ') uip++;
    strcpy(oneline,uip); 
    if (strstr(uip, ".") > 0)  {
      /* advance to next sentence:*/
      uip = strstr(uip, ".")+1;
      /* strip out first sentence:*/
      *strstr(oneline, ".") = 0;
    }   
    if (oneline != NULL && strlen(oneline) > 0) {
      /* Robot Respose:*/
      cleaner(oneline);
      strcat(output, respond2( oneline ));   

      if (logging) {
        fprintf(logfile, "%s'%s' %s/%s %s/%s %s\n",
         ltimestr,getvar("name",NULL),getvar("age",NULL),
         getvar("personality","normal"),getvar("gender",NULL),
         getvar("location",NULL),getvar("ip",NULL));
        fprintf(logfile, "Client: %s\n",oneline);
        /* DW says we sould strip HTML out here (shrug) */
        fprintf(logfile, "Robot: %s\n",output);
      }

      strcat(output,"  ");
    }
  } while (strstr(uip, ".") > 0); /* any more sentences? */
  delete(oneline);
  delete(scrap);
  strcpy(pbuffer, output);
  strcat(pbuffer,"<br>");
  delete(output);
  print(pbuffer);
  return(pbuffer);
}

void savevars( char* varfile ) {
  FILE* vars = fopen(varfile,"w");
  var_array *conductor =root_var;
  /* first we print out that. */
  fprintf(vars,"%s\n",that);
  /* we print out each variable. */
  while(conductor != NULL) {
    if (conductor->def == 1)
    fprintf(vars,"%i %s=*%s\n",conductor->protect,conductor->varname,conductor->value);
    else
    fprintf(vars,"%i %s=%s\n",conductor->protect,conductor->varname,conductor->value);
    conductor= conductor->next;
  }
  fclose(vars);
}

void loadvars( char* varfile ) {
  char *scrap = new char[MAX_LINE_SIZE];
  char *scrap2 = new char[MAX_LINE_SIZE];
  var_array *conductor = root_var;
  /* first wipe out our exsisting variables */
  while (conductor != NULL) {
    conductor=conductor->next;
    delete(root_var->varname);
    delete(root_var->value);
    delete(root_var);
    root_var = conductor;
  }
 
  /* next we see if varfile exsists. */
  FILE *vars = fopen(varfile,"r");
  if(vars){
    /* if so, we load up variables from it into memory. */
    /* first we grab that */
    fgets(that,MAX_LINE_SIZE,vars);
    /* remove the EOL char */
    replace(that,"\n","");
    /* now we load up our vars. */
    while(!feof(vars)) {
      fgets(scrap,MAX_LINE_SIZE,vars);
      replace(scrap,"\n","");
      if(!(strstr(scrap,"="))) continue;
      strcpy(scrap2,scrap + 2); 
      scrap[1] = 0;
      int protect = atoi(scrap);      
      strcpy(scrap,scrap2);
      *strchr(scrap,'=') = 0;
      char *varname = scrap;
      char *value = strstr(scrap2,"=") +1;
      setvar(protect,varname,value);
    }/*end while */
    fclose(vars);
  }
  else {
    /*if not, we load up defvars.txt into our varfile, and into memory.
    betcha thought this was gonna be hard :) */
    loadvars(datafile);
    savevars(varfile);
  }
  delete(scrap);
  delete(scrap2);
}

int initialize(char *init_file){
  /* scrap variables for our init file */
  char *aimlscrap, *logscrap, *patternscrap, *templatescrap;
  char *errorscrap, *gossipscrap;
  char *personscrap, *personfscrap, *person2scrap;
  char *substitutescrap;

  /* Buffers for '*', the last response and the print routine */
  star = new char[MAX_LINE_SIZE];
  that = new char[MAX_LINE_SIZE];
  best = new char[MAX_LINE_SIZE];
  pbuffer = new char[MAX_LINE_SIZE];

  root_var=NULL;

  char *epos;

  datafile = new char[MAX_FILENAME_SIZE];

  aimlscrap = new char[MAX_FILENAME_SIZE];
  logscrap = new char[MAX_FILENAME_SIZE];
  patternscrap = new char[MAX_FILENAME_SIZE];
  templatescrap = new char[MAX_FILENAME_SIZE];
  unsortedfile = new char[MAX_FILENAME_SIZE];
  errorscrap = new char[MAX_FILENAME_SIZE];
  substitutescrap = new char[MAX_FILENAME_SIZE];
  personscrap = new char[MAX_FILENAME_SIZE];
  person2scrap = new char[MAX_FILENAME_SIZE];
  personfscrap = new char[MAX_FILENAME_SIZE];
  gossipscrap = new char[MAX_FILENAME_SIZE];

  /*start the radomizer for the <random> tags */
  srand(time(0));

  /* load up the init file. */
  FILE* init = fopen( init_file, "r" );
  if (!init) {
    printf("Init file not found - %s aborting\n",init_file);
    return(1);
  }
  while(!feof(init)) {
    //we use pbuffer as a scrap variable.
    fgets(pbuffer, MAX_LINE_SIZE, init);
    //eliminate comments.
    if(strchr(pbuffer,'#')) *strchr(pbuffer,'#') = 0;
    //delete all spaces
    while(replace(pbuffer," ",""));
    //remove EOL chars
    while(replace(pbuffer,"\n",""));
      epos = strchr(pbuffer,'=');
/*    if (!(epos = strchr(pbuffer,'=')) || (*pbuffer != 0)){
      printf("Error in it file aborting\n");
      return(1); 
    } */
    epos++;
    /* these really should be strncmp()'s */
    if(strstr(pbuffer,"aimlfile")) strcpy(aimlscrap,epos);
    if(strstr(pbuffer,"logfile")) strcpy(logscrap,epos);
    if(strstr(pbuffer,"datafile")) strcpy(datafile,epos);
    if(strstr(pbuffer,"personfile")) strcpy(personscrap,epos);
    if(strstr(pbuffer,"person2file")) strcpy(person2scrap,epos);
    if(strstr(pbuffer,"personffile")) strcpy(personfscrap,epos);
    if(strstr(pbuffer,"substitutefile")) strcpy(substitutescrap,epos);
    if(strstr(pbuffer,"patternfile")) strcpy(patternscrap,epos);
    if(strstr(pbuffer,"templatefile")) strcpy(templatescrap,epos);
    if(strstr(pbuffer,"unsortedfile")) strcpy(unsortedfile,epos);
    if(strstr(pbuffer,"gossipfile")) strcpy(gossipscrap,epos);
    if(strstr(pbuffer,"errorfile")) strcpy(errorscrap,epos);

    if (strstr(pbuffer,"logging")) {
      if(*epos == 'y' || *epos == 'Y' || *epos == '1')
        logging = 1;
      else 
        logging = 0;
    }

    if (strstr(pbuffer,"log_error")) {
      if(*epos == 'y' || *epos == 'Y' || *epos == '1')
        log_error = 1; 
      else 
        log_error = 0;
    }
  } /* end while feof(init) */

  /* open our logfile and errorlog */        
  if (log_error) {
    errorfile = fopen(errorscrap,"a");
    if (!errorfile) {
      printf("Error opening errorfile\n");
      return(1);
    }
  }
  if (logging) {
    logfile = fopen(logscrap, "a");
    if(!logfile) {
      alierror("Error opening logfile\n");
      return(1);
    } 
  }
  gossipfile = fopen(gossipscrap,"a");
    if(!gossipfile) {
      alierror("Error opening gossipfile\n");
      return(1);
    } 
  personfile = fopen(personscrap,"r");
    if(!personfile) {
      alierror("Error opening personfile\n");
      return(1);
    } 
  person2file = fopen(person2scrap,"r");
    if(!person2file) {
      alierror("Error opening person2file\n");
      return(1);
    } 
  personffile= fopen(personfscrap,"r");
    if(!personffile) {
      alierror("Error opening personffile\n");
      return(1);
    } 
  substitutefile = fopen(substitutescrap,"r");
    if(!substitutefile) {
      alierror("Error opening substitutefile\n");
      return(1);
    } 
  
  //check to see if it needs to create new patterns.txt and templates.txt

  patternfile = fopen(patternscrap,"r");
  templatefile = fopen(templatescrap,"r");

  if (!patternfile || !templatefile){
    if (patternfile) fclose(patternfile); 
    if (templatefile) fclose(templatefile);
    patternfile = fopen(patternscrap,"w");
      if(!patternfile) {
        alierror("Error opening patternfile\n");
        return(1);
      } 
    templatefile = fopen(templatescrap,"w");
    if(!templatefile) {
      alierror("Error opening templatefile\n");
      return(1);
    } 
    aimlfile = fopen(aimlscrap,"r");
    if(!aimlfile) {
      alierror("Error opening aimlfile\n");
      return(1);
    } 
    if (!prepare()) return (1);
    fclose(aimlfile);
    fclose(patternfile);
    fclose(templatefile);
    patternfile = fopen(patternscrap,"r");
    if(!patternfile) {
      alierror("Error opening patternfile\n");
      return(1);
    } 
    templatefile = fopen(templatescrap,"r");
    if(!templatefile) {
      alierror("Error opening templatefile\n");
      return(1);
    } 
  }
  else{
    /* ok, this part really should be saved to a file upon preparation */
    char last = 0;
    while (!feof( patternfile )){
      fpos_t line;
      fgetpos( patternfile, &line );
      fgets( pbuffer, MAX_LINE_SIZE, patternfile );
      char first = pbuffer[0];
      if ((first < 'A') || (first > 'Z')) first = 'Z' + 1;
      if (first != last){
        last = first;
        if ((last < 'A') || (last > 'Z')) last = 'Z' + 1;
        chunk[last - 'A'] = line;
      }
    }
  }
  delete(aimlscrap);
  delete(logscrap);
  delete(patternscrap);
  delete(templatescrap);
  delete(errorscrap);
  delete(substitutescrap);
  delete(personscrap);
  delete(person2scrap);
  delete(personfscrap);
  delete(gossipscrap);
  return(0);
}

void deinitialize() {
  if (logging) fclose(logfile);
  if (log_error) fclose(errorfile);
  delete(star);
  delete(that);
  delete(best);
  delete(pbuffer);
  delete(datafile);
  delete(unsortedfile);
  fclose(errorfile);  
  fclose(substitutefile);
  fclose(templatefile);
  fclose(patternfile);
  fclose(gossipfile);
  fclose(personfile);
  fclose(personffile);
  fclose(person2file);
}
