#!/usr/bin/perl
package NSH;

use Cwd;
use Text::ParseWords;
sub processInput {
my $input = shift;
chomp $input;
@args = eval { &shellwords($input) };
if ($@) {
    $error = $@;
    $error =~ /^# ([^\n]*)\n/;
    $error = ($1 || $error);
    doError($error, "<STDIN>");
    next;
}
$command = shift @args;
&doCommand($command, [@args]);
}
sub doCommand {
    my $command = shift;
    if(defined &{ "NSH::Commands::$command" }) {
	&{ "NSH::Commands::$command" }(shift);
    } elsif (defined($com = NSH::Commands::_load($command))) {
	&{ "NSH::Commands::$command" }(shift);
    } else {
	print "ERROR: No such command: $command.\n";
    }
}
sub doError {
    my ($error, $loc) = @_;
    print "<PRE>\n";
    print "+" . ("=" x 70) . "+\n";
    printf("| %-68s |\n", "ERROR: " . $error);
    printf("| %-68s |\n", "\@ " . localtime() . "; " . $loc);
    print "+" . ("=" x 70) . "+\n";
    print "</PRE>\n";
}

package NSH::Commands;
sub quit {
    $NSH::DONE = 1;
}
sub cd {
    $dir = $_[0]->[0];
    if (($^O =~ /MacOS/) && ($dir eq "..")){ $dir = "::" };
    chdir $dir || chdir ($dir . (($^O =~ /MacOS/) ? ":" : "/")) || NSH::doError("No such directory!", "<STDIN>");
}
sub _load {
    $com = shift;
    opendir(EXT, "ext");
    @exts = readdir(EXT);
    closedir(EXT);
    chomp @exts;
    my @possibilities = map { /^$com$/i ? ($_) : () } @exts;
    my @poss = @possibilities;
    #foreach (@possibilities) {
    #    if ($_) {
    #        push @poss, $_;
    #    }
    #}
    
    if (@poss == 1) {
        my $comName = $poss[0];
	#\&{"NSH::Commands::$comName"} = undef;
        require "ext/$comName";
        return 1;
        
    }
    
    return undef;
}
sub init {
    NSH::Utilities::setPref("bot.style", "alice");
}

sub bot {
    my $args = shift;
	if ($args->[0]) {
		NSH::Utilities::setPref("bot.style", $args->[0]);
		print "You will now talk with: $args->[0].";
	} else {
		print "You are currently talking with ", NSH::Utilities::getPref("bot.style");
	}
}

package NSH::Utilities;

sub getPref {
    my $pref = shift;
    if ($pref =~ /\//) {
	my ($user, $preff) = split (/\//, $pref);
	$user =~ s/ //g;
	my ($pack, $pname) = split (/\./, $preff);
	return _slurp($ENV{UHOME} . "../$user/$pack.pref", $pname);
    } else {
	my ($pack, $pname) = split (/\./, $pref);
	return _slurp($ENV{UHOME} . $pack . ".pref", $pname);
    }
}
sub setPref {
	my $pref = shift;
	my $to = shift;
	my $user;
	print $pref, "\n";
	if ($pref =~ /\//) {
	    ($user, $pref) = split (/\//, $pref);
	    $user =~ s/ //g;
	}
	#print "\$user = $user, \$pref = $pref\n";
	my ($pack, $pname) = split (/\./, $pref);
	#print "\$pack = $pack, \$pname = $pname\n";
	my @curprefs = _slurp($ENV{UHOME} . ($user ? "../$user/" : "") . $pack . ".pref");
	my $foundit = 0;
	foreach (@curprefs) {
		if($_ =~ /^$pname\|/) {
			$_ = "$pname\|$to\n";
			$foundit = 1;
		}
	}
	unless ($foundit) {
		push(@curprefs, "$pname\|$to\n");
	}
	#print "Opening ",$ENV{UHOME} . ($user ? "../$user/" : "") . $pack . ".pref\n";
	open(PREFFILE, ">".$ENV{UHOME} . ($user ? "../$user/" : "") . $pack . ".pref");
	print PREFFILE @curprefs;
	#print @curprefs;
	close(PREFFILE);
}
sub _slurp {
    my $file = shift;
    my $key = shift;
    if ($key) {
        my $contents = _slurp($file);
        $contents =~ /$key\|([^\n]*)\n/;
        return $1;
    } else {
	if (wantarray) {
	    $file =~ tr/\//\:/ if ($^O =~ /MacOS/);
	    my $error = 0;
	    if (!-e $file) {return ();}
	    open(IN, $file) or $error = $!;
	    if ($error) {
	        $file = ":" . $file if ($^O =~ /MacOS/); # Relative paths on macs start with ":"
		open(IN, $file) and $error = 0;
	    }
	    die "Couldn't open $file: $error" if $error;
	    @lines = <IN>;
	    return @lines;
	} else {
	    my @lines = _slurp($file);
	    return join("", @lines);
	}
    }
}

1;